@@ -1,5 +1,5 @@
 /*
- * Copyright(C) 2011 Canonical Ltd.
+ * Copyright(C) 2013 Canonical Ltd.
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *  You should have received a copy of the GNU General Public License
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *
- * Authored by Ken VaDine <ken.vandine@canonical.com>
+ * Authored by Ken VanDine <ken.vandine@canonical.com>
  */
 
 #include <glib.h>
@@ -35,7 +35,9 @@ static MessagingMenuApp *mmapp;
 GHashTable *dispatch_ops;
 static gint n_sources;
 UnityLauncherEntry *launcher = NULL;
-GList *conns;
+GHashTable *conns; /* connection object -> set of mmapp ids for that connection */
+TpBaseClient *approver;
+TpBaseClient *observer;
 
 static void update_launcher (gint count)
 {
@@ -55,30 +57,51 @@ static void update_launcher (gint count)
 	}
 }
 
+void
+remember_id_for_connection (TpConnection *con,
+			    const gchar  *id)
+{
+	GHashTable *ids;
+
+	ids = g_hash_table_lookup (conns, con);
+	if (ids != NULL)
+		g_hash_table_add (ids, g_strdup (id));
+	else
+		g_warning ("%s: never seen chanel '%s' before", G_STRFUNC, id);
+}
+
 static gchar *
 channel_id (TpChannel *channel)
 {
 	TpConnection *con;
+	gchar *id;
 
-	con = tp_channel_borrow_connection (channel);
+	con = tp_channel_get_connection (channel);
+	id = g_strconcat ("channel:",
+			  tp_proxy_get_object_path (TP_PROXY (con)), ":",
+			  tp_proxy_get_object_path (TP_PROXY (channel)),
+			  NULL);
 
-	return g_strconcat ("channel:",
-			    tp_proxy_get_object_path (TP_PROXY (con)), ":",
-			    tp_proxy_get_object_path (TP_PROXY (channel)),
-			    NULL);
+	remember_id_for_connection (con, id);
+
+	return id;
 }
 
 static gchar *
 contact_id (TpContact *contact)
 {
 	TpConnection *con;
+	gchar *id;
 
 	con = tp_contact_get_connection (contact);
+	id = g_strconcat ("contact:",
+			  tp_proxy_get_object_path (TP_PROXY (con)), ":",
+			  tp_contact_get_identifier (contact),
+			  NULL);
 
-	return g_strconcat ("contact:",
-			    tp_proxy_get_object_path (TP_PROXY (con)), ":",
-			    tp_contact_get_identifier (contact),
-			    NULL);
+	remember_id_for_connection (con, id);
+
+	return id;
 }
 
 static void
@@ -87,7 +110,7 @@ dispatch_op_finished_cb (TpProxy *self, guint domain, gint code, gchar *message,
 	TpChannelDispatchOperation *dispatch_op = TP_CHANNEL_DISPATCH_OPERATION (self);
 	GPtrArray *channels;
 
-	channels = tp_channel_dispatch_operation_borrow_channels (dispatch_op);
+	channels = tp_channel_dispatch_operation_get_channels (dispatch_op);
 	if (channels != NULL && channels->len)
 	{
 		TpChannel *channel;
@@ -96,7 +119,8 @@ dispatch_op_finished_cb (TpProxy *self, guint domain, gint code, gchar *message,
 		channel = g_ptr_array_index (channels, 0);
 		id = channel_id (channel);
 
-		messaging_menu_app_remove_source (mmapp, id);
+		if (mmapp)
+			messaging_menu_app_remove_source (mmapp, id);
 		g_hash_table_remove (dispatch_ops, id);
 		update_launcher (n_sources--);
 
@@ -113,7 +137,7 @@ handle_with_cb (GObject *self, GAsyncResult *result, gpointer user_data)
 	GPtrArray *channels;
 	gchar *id;
 
-	channels = tp_channel_dispatch_operation_borrow_channels (dispatch_op);
+	channels = tp_channel_dispatch_operation_get_channels (dispatch_op);
 	channel = g_ptr_array_index (channels, 0);
 	id = channel_id (channel);
 
@@ -123,7 +147,8 @@ handle_with_cb (GObject *self, GAsyncResult *result, gpointer user_data)
 		{
 			g_warning ("Failed to handle operations: %s\n", error->message);
 			g_error_free (error);
-			messaging_menu_app_remove_source (mmapp, id);
+			if (mmapp)
+				messaging_menu_app_remove_source (mmapp, id);
 			update_launcher (n_sources--);
 		}
 	}
@@ -243,8 +268,6 @@ handle_contacts_subscription_add_indicator_cb (GObject     *object,
 {
 	TpContact *contact;
 	GtkWidget *dialog;
-	GtkWidget *button;
-	gchar *id;
 	const gchar *alias = "";
 	gchar *title;
 	const gchar *message;
@@ -303,18 +326,22 @@ handle_contacts_subscription_add_indicator_cb (GObject     *object,
 
 	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_ACCEPT);
 
-	button = gtk_dialog_add_button (GTK_DIALOG (dialog), _("_Decline"), GTK_RESPONSE_REJECT);
+	gtk_dialog_add_button (GTK_DIALOG (dialog), _("_Decline"), GTK_RESPONSE_REJECT);
 
-	button = gtk_dialog_add_button (GTK_DIALOG (dialog), _("_Accept"), GTK_RESPONSE_ACCEPT);
+	gtk_dialog_add_button (GTK_DIALOG (dialog), _("_Accept"), GTK_RESPONSE_ACCEPT);
 
 	g_signal_connect (dialog, "response", G_CALLBACK (contact_add_response_cb), contact);
 
 	gtk_widget_show (GTK_WIDGET (dialog));
 
-	id = contact_id (contact);
-	messaging_menu_app_remove_source (mmapp, id);
+	if (mmapp) {
+		gchar *id;
 
-	g_free (id);
+		id = contact_id (contact);
+		messaging_menu_app_remove_source (mmapp, id);
+
+		g_free (id);
+	}
 }
 
 static TpConnection *
@@ -392,9 +419,11 @@ message_source_activated (MessagingMenuApp *app,
 			  const gchar      *id,
 			  gpointer          user_data)
 {
-	TpAccountManager *tp_am = user_data;
+	TpAccountManager *tp_am;
 	gchar **sections;
 
+	tp_am = tp_account_manager_dup ();
+
 	/* message source ids contain the type of source, the
 	 * tpconnection's object path, and the object path of the actual
 	 * source object, separated by colons */
@@ -451,33 +480,29 @@ message_source_activated (MessagingMenuApp *app,
 
 out:
 	g_strfreev (sections);
+	g_object_unref (tp_am);
 }
 
-static TpContact *
-get_contact_from_channel (TpChannel * channel)
-{
-	g_return_val_if_fail (TP_IS_CHANNEL (channel), NULL);
-	TpContact *contact = NULL;
-	if (tp_channel_get_requested (TP_CHANNEL (channel)))
-		contact = tp_channel_get_target_contact (TP_CHANNEL (channel));
-	else
-		contact = tp_channel_get_initiator_contact (TP_CHANNEL (channel));
-	return TP_CONTACT (contact);
-}
-	
-
 static void
-handle_contacts_pending_add_indicator_cb (TpConnection *connection, guint n_contacts, TpContact * const *contacts, guint n_failed, const TpHandle *failed, const GError *err, gpointer user_data, GObject *weak_object)
+upgrade_contacts_cb (GObject *source_object, GAsyncResult *result, gpointer user_data)
 {
-	g_debug ("handle_contacts_pending_add_indicator_cb");
-	TpChannel *channel = TP_CHANNEL (user_data);
-	gchar *id;
+	TpChannel *channel = user_data;
+	GPtrArray *contacts;
+	GError *error = NULL;
 	const gchar *alias = "";
 	GFile *avatar_file;
 	GIcon *avatar = NULL;
 	TpContact *contact = NULL;
 
-	contact = get_contact_from_channel (channel);
+	if (!tp_connection_upgrade_contacts_finish (TP_CONNECTION (source_object), result, &contacts, &error)) {
+		g_warning ("unable to fetch contact information: %s", error->message);
+		g_error_free (error);
+		return;
+	}
+
+	g_return_if_fail (contacts->len == 1);
+
+	contact = g_ptr_array_index (contacts, 0);
 
 	if (contact)
 	{
@@ -485,21 +510,23 @@ handle_contacts_pending_add_indicator_cb (TpConnection *connection, guint n_cont
 		avatar_file = tp_contact_get_avatar_file (contact);
 		avatar = avatar_file ? g_file_icon_new (avatar_file) : NULL;
 		g_debug ("handle_contacts_pending_add_indicator_cb HAS CONTACT alias: %s", alias);
-	} else
-	{
-		alias = tp_channel_get_initiator_identifier (channel);
-		g_debug ("handle_contacts_pending_add_indicator_cb NO CONTACT alias: %s", alias);
 	}
 	g_debug ("handle_contacts_pending_add_indicator_cb alias: %s", alias);
 
-	id = channel_id (channel);
-	messaging_menu_app_append_source (mmapp, id, avatar, alias);
-	messaging_menu_app_draw_attention (mmapp, id);
+	if (mmapp) {
+		gchar *id;
+
+		id = channel_id (channel);
+		messaging_menu_app_append_source (mmapp, id, avatar, alias);
+		messaging_menu_app_draw_attention (mmapp, id);
+
+		g_free (id);
+	}
 
 	update_launcher (++n_sources);
 
 	g_clear_object (&avatar);
-	g_free (id);
+	g_ptr_array_free (contacts, TRUE);
 }
 
 gboolean check_pending_messages (gpointer data)
@@ -507,15 +534,14 @@ gboolean check_pending_messages (gpointer data)
 	TpTextChannel *channel = TP_TEXT_CHANNEL (data);
 	TpConnection *connection;
 	GList *pending;
-	gboolean has_indicator = FALSE;
-	TpHandle handle;
+	TpContact *contact;
 
 	static TpContactFeature features[] = {
 		TP_CONTACT_FEATURE_ALIAS,
 		TP_CONTACT_FEATURE_AVATAR_DATA,
 	};
 
-	pending = tp_text_channel_get_pending_messages (TP_TEXT_CHANNEL (channel));
+	pending = tp_text_channel_dup_pending_messages (TP_TEXT_CHANNEL (channel));
 	g_debug ("pending: %u", g_list_length (pending));
 	if (g_list_length (pending) > 0)
 	{
@@ -523,30 +549,35 @@ gboolean check_pending_messages (gpointer data)
 
 		id = channel_id (TP_CHANNEL (channel));
 
-		if (!messaging_menu_app_has_source (mmapp, id))
+		if (mmapp && !messaging_menu_app_has_source (mmapp, id))
 		{
-			connection = tp_channel_borrow_connection (TP_CHANNEL (channel));
-			handle = tp_channel_get_initiator_handle (TP_CHANNEL (channel));
-			tp_connection_get_contacts_by_handle (connection, 1, &handle, 2, features, handle_contacts_pending_add_indicator_cb, channel, NULL, NULL);
+			connection = tp_channel_get_connection (TP_CHANNEL (channel));
+			contact = tp_channel_get_initiator_contact (TP_CHANNEL (channel));
+			if (contact)
+				tp_connection_upgrade_contacts_async (connection, 1, &contact, 2, features, upgrade_contacts_cb, channel);
 		}
 
 		g_free (id);
 	}
-        g_list_free (pending);
+	g_list_free_full (pending, g_object_unref);
 	return FALSE;
 }
 
 static void
 pending_message_removed_cb (TpTextChannel *channel, TpMessage *message, gpointer user_data)
 {
-	gchar *id;
 	g_debug ("pending_message_removed_cb: %s", tp_channel_get_identifier (TP_CHANNEL (channel)));
 
-	id = channel_id (TP_CHANNEL (channel));
-	messaging_menu_app_remove_source (mmapp, id);
-	update_launcher (--n_sources);
+	if (mmapp) {
+		gchar *id;
 
-	g_free (id);
+		id = channel_id (TP_CHANNEL (channel));
+		messaging_menu_app_remove_source (mmapp, id);
+
+		g_free (id);
+	}
+
+	update_launcher (--n_sources);
 }
 
 static void
@@ -557,53 +588,59 @@ message_received_cb (TpTextChannel *channel, TpMessage *message, gpointer user_d
 }
 
 static void
-handle_contacts_add_indicator_cb (TpConnection *connection, guint n_contacts, TpContact * const *contacts, guint n_failed, const TpHandle *failed, const GError *err, gpointer user_data, GObject *weak_object)
+handle_contacts_add_indicator_cb (GObject *source_object, GAsyncResult *result, gpointer user_data)
 {
+	TpChannelDispatchOperation *dispatch_op = user_data;
+	GPtrArray *contacts;
+	GError *error = NULL;
 	gchar *id;
-	
-	TpChannelDispatchOperation *dispatch_op = TP_CHANNEL_DISPATCH_OPERATION (user_data);
-
 	TpChannel *channel;
 	GPtrArray *channels;
-	channels = tp_channel_dispatch_operation_borrow_channels (dispatch_op);
+	channels = tp_channel_dispatch_operation_get_channels (dispatch_op);
 	channel = g_ptr_array_index (channels, 0);
 
+	if (!tp_connection_upgrade_contacts_finish (TP_CONNECTION (source_object), result, &contacts, &error)) {
+		g_warning ("unable to fetch contact information: %s", error->message);
+		g_error_free (error);
+		return;
+	}
+
+	g_return_if_fail (contacts->len == 1);
+
 	id = channel_id (channel);
 
-	if (messaging_menu_app_has_source (mmapp, id))
-	{
-		messaging_menu_app_set_source_time (mmapp, id, g_get_real_time ());
-	}
-	else
-	{
-		TpContact *contact;
-		const gchar *alias = "";
-		GFile *avatar_file;
-		GIcon *avatar;
-
-		contact = get_contact_from_channel (channel);
-		if (contact)
+	if (mmapp) {
+		if (messaging_menu_app_has_source (mmapp, id))
 		{
+			messaging_menu_app_set_source_time (mmapp, id, g_get_real_time ());
+		}
+		else
+		{
+			TpContact *contact;
+			const gchar *alias = "";
+			GFile *avatar_file;
+			GIcon *avatar;
+
+			contact = g_ptr_array_index (contacts, 0);
 			alias = tp_contact_get_alias (contact);
 			avatar_file = tp_contact_get_avatar_file (contact);
 			avatar = avatar_file ? g_file_icon_new (avatar_file) : NULL;
-		} else
-		{
-			alias = tp_channel_get_initiator_identifier (channel);
+
+			messaging_menu_app_append_source (mmapp, id, avatar, alias);
+
+			g_clear_object (&avatar);
 		}
 
-		messaging_menu_app_append_source (mmapp, id, avatar, alias);
-		update_launcher (++n_sources);
-
-		g_clear_object (&avatar);
+		messaging_menu_app_draw_attention (mmapp, id);
 	}
 
-	messaging_menu_app_draw_attention (mmapp, id);
+	update_launcher (++n_sources);
 
 	g_hash_table_insert (dispatch_ops, g_strdup (id), dispatch_op);
 	g_signal_connect (dispatch_op, "invalidated", G_CALLBACK (dispatch_op_finished_cb), NULL);
 
 	g_free (id);
+	g_ptr_array_free (contacts, TRUE);
 }
 
 static void
@@ -613,7 +650,7 @@ add_dispatch_operation_cb (TpSimpleApprover *self, TpAccount *account, TpConnect
 {
 	GList *itr;
 	TpChannel *channel;
-	TpHandle handle;
+	TpContact *contact;
 
 	static TpContactFeature features[] = {
 		TP_CONTACT_FEATURE_ALIAS,
@@ -624,8 +661,8 @@ add_dispatch_operation_cb (TpSimpleApprover *self, TpAccount *account, TpConnect
 	for (itr = channels; itr != NULL; itr = g_list_next (itr))
 	{
 		channel = itr->data;
-		handle = tp_channel_get_initiator_handle (channel);
-		tp_connection_get_contacts_by_handle (connection, 1, &handle, 2, features, handle_contacts_add_indicator_cb, dispatch_op, NULL, NULL);
+		contact = tp_channel_get_initiator_contact (channel);
+		tp_connection_upgrade_contacts_async (connection, 1, &contact, 2, features, handle_contacts_add_indicator_cb, dispatch_op);
 	}
 	g_list_free(itr);
 	tp_add_dispatch_operation_context_accept (context);
@@ -664,8 +701,7 @@ static void
 conn_invalidated_cb (TpConnection *conn, guint domain, gint code, gchar *message, gpointer user_data)
 {
 	g_debug ("conn_invalidated_cb");
-	conns = g_list_remove (conns, conn);
-	g_object_unref (conn);
+	g_hash_table_remove (conns, conn);
 }
 
 static void
@@ -690,22 +726,24 @@ contact_list_changed_cb (TpConnection *conn, GPtrArray *added, GPtrArray *remove
 
 				id = contact_id (contact);
 
-				if (messaging_menu_app_has_source (mmapp, id))
-				{
-					messaging_menu_app_set_source_time (mmapp, id, g_get_real_time ());
-				}
-				else
-				{
-					alias = tp_contact_get_alias (contact);
-					avatar_file = tp_contact_get_avatar_file (contact);
-					avatar = avatar_file ? g_file_icon_new (avatar_file) : NULL;
+				if (mmapp) {
+					if (messaging_menu_app_has_source (mmapp, id))
+					{
+						messaging_menu_app_set_source_time (mmapp, id, g_get_real_time ());
+					}
+					else
+					{
+						alias = tp_contact_get_alias (contact);
+						avatar_file = tp_contact_get_avatar_file (contact);
+						avatar = avatar_file ? g_file_icon_new (avatar_file) : NULL;
 
-					messaging_menu_app_append_source (mmapp, id, avatar, alias);
-					messaging_menu_app_draw_attention (mmapp, id);
+						messaging_menu_app_append_source (mmapp, id, avatar, alias);
+						messaging_menu_app_draw_attention (mmapp, id);
 
-					n_sources++;
+						n_sources++;
 
-					g_clear_object (&avatar);
+						g_clear_object (&avatar);
+					}
 				}
 
 				g_debug ("contact_list_changed_cb: adding %s", alias);
@@ -725,7 +763,8 @@ contact_list_changed_cb (TpConnection *conn, GPtrArray *added, GPtrArray *remove
 			contact_remove_from_contact_list (contact);
 		}
 
-		messaging_menu_app_remove_source (mmapp, tp_contact_get_identifier (contact));
+		if (mmapp)
+			messaging_menu_app_remove_source (mmapp, tp_contact_get_identifier (contact));
 		n_sources--;
 	}
 
@@ -741,12 +780,13 @@ check_account (TpAccount *account, TpAccountManager *manager)
 	if (connection == NULL)
 		return;
 
-	if (g_list_find (conns, connection) != NULL)
+	if (g_hash_table_contains (conns, connection))
 		return;
 
-	g_debug ("account_manager_prepared_cb %s", tp_connection_get_connection_manager_name (connection));
+	g_debug ("account_manager_prepared_cb %s", tp_connection_get_cm_name (connection));
 
-	conns = g_list_prepend (conns, g_object_ref (connection));
+	g_hash_table_insert (conns, g_object_ref (connection),
+			     g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL));
 
 	tp_g_signal_connect_object (connection, "contact-list-changed",
 		G_CALLBACK (contact_list_changed_cb), manager, 0);
@@ -768,6 +808,8 @@ add_account (TpAccount *account, TpAccountManager *manager)
 	g_debug ("add_account");
 	tp_g_signal_connect_object (account, "notify::connection",
 		G_CALLBACK (account_conn_changed_cb), manager, 0);
+	if (mmapp != NULL)
+		messaging_menu_app_register (mmapp);
 	check_account (account, manager);
 }
 
@@ -775,7 +817,6 @@ static void
 account_enabled_cb (TpAccountManager *manager, TpAccount *account, gboolean valid, gpointer user_data)
 {
 	g_debug ("account_enabled_cb");
-
 	add_account (account, manager);
 }
 
@@ -783,7 +824,6 @@ static void
 account_manager_prepared_cb (GObject *object, GAsyncResult *res, gpointer user_data)
 {
 	TpAccountManager *manager = TP_ACCOUNT_MANAGER (object);
-        TpAccountManager *account_manager = TP_ACCOUNT_MANAGER (user_data);
 	GList *accounts;
 	GError *error = NULL;
 
@@ -794,17 +834,21 @@ account_manager_prepared_cb (GObject *object, GAsyncResult *res, gpointer user_d
 		g_print ("Error preparing AM: %s\n", error->message);
 	}
 
-	for (accounts = tp_account_manager_get_valid_accounts (manager); accounts != NULL; accounts = g_list_delete_link (accounts, accounts))
+	accounts = tp_account_manager_dup_valid_accounts (manager);
+	while (accounts != NULL)
 	{
 		TpAccount *account = accounts->data;
 		g_debug ("account_manager_prepared_cb %s", tp_account_get_path_suffix (account));
-		guint i;
 
 		add_account (account, manager);
+
+		accounts = accounts->next;
 	}
 
 	tp_g_signal_connect_object (manager, "account-validity-changed",
 		G_CALLBACK (account_enabled_cb), manager, 0);
+
+	g_list_free_full (accounts, g_object_unref);
 }
 
 void
@@ -835,12 +879,9 @@ most_available_presence_changed_cb (TpAccountManager *manager, guint presence, g
 }
 
 static TpBaseClient *
-approver_setup (TpAccountManager *tp_am)
+approver_setup (TpAccountManager *tp_am, GError **error)
 {
-	TpBaseClient *approver;
-	GError *error = NULL;
-
-	g_return_if_fail (TP_IS_ACCOUNT_MANAGER (tp_am));
+	g_return_val_if_fail (TP_IS_ACCOUNT_MANAGER (tp_am), NULL);
 
 	approver = tp_simple_approver_new_with_am (tp_am, "IndicatorApprover",
 				FALSE, add_dispatch_operation_cb, NULL, NULL);
@@ -857,27 +898,18 @@ approver_setup (TpAccountManager *tp_am)
 		TP_PROP_CHANNEL_TARGET_HANDLE_TYPE, G_TYPE_UINT, TP_HANDLE_TYPE_ROOM,
 		NULL));
 
-	if (!tp_base_client_register (approver, &error))
+	if (!tp_base_client_register (approver, error))
 	{
-                if (error)
-		{
-			g_warning ("Failed to register the approver: %s", error->message);
-			g_error_free (error);
-		}
-		g_object_unref (tp_am);
-		g_object_unref (approver);
+		g_clear_object (&approver);
 		return NULL;
 	}
 	return approver;
 }
 
 static TpBaseClient *
-observer_setup (TpAccountManager *tp_am)
+observer_setup (TpAccountManager *tp_am, GError **error)
 {
-	TpBaseClient *observer;
-	GError *error = NULL;
-
-	g_return_if_fail (TP_IS_ACCOUNT_MANAGER (tp_am));
+	g_return_val_if_fail (TP_IS_ACCOUNT_MANAGER (tp_am), NULL);
 
 	observer = tp_simple_observer_new_with_am (tp_am, TRUE, "IndicatorObserver",
 				FALSE, observer_observe_channels_cb, NULL, NULL);
@@ -888,15 +920,9 @@ observer_setup (TpAccountManager *tp_am)
 		TP_PROP_CHANNEL_TARGET_HANDLE_TYPE, G_TYPE_UINT, TP_HANDLE_TYPE_CONTACT,
 		NULL));
 
-	if (!tp_base_client_register (observer, &error))
+	if (!tp_base_client_register (observer, error))
 	{
-		if (error)
-		{
-			g_warning ("Failed to register the observer: %s", error->message);
-			g_error_free (error);
-		}
-		g_object_unref (tp_am);
-		g_object_unref (observer);
+		g_clear_object (&observer);
 		return NULL;
 	}
 	return observer;
@@ -911,6 +937,9 @@ presence_changed (TpAccountManager *manager,
 {
 	MessagingMenuStatus mm_status;
 
+	if (mmapp == NULL)
+		return;
+
 	switch (presence)
 	{
 	case TP_CONNECTION_PRESENCE_TYPE_AVAILABLE:
@@ -946,9 +975,11 @@ messaging_menu_status_changed (MessagingMenuApp    *mmapp,
 			       MessagingMenuStatus mm_status,
 			       gpointer            user_data)
 {
-	TpAccountManager *tp_am = user_data;
+	TpAccountManager *tp_am;
 	TpConnectionPresenceType status;
 
+	tp_am = tp_account_manager_dup ();
+
 	switch (mm_status)
 	{
 	case MESSAGING_MENU_STATUS_AVAILABLE:
@@ -976,60 +1007,193 @@ messaging_menu_status_changed (MessagingMenuApp    *mmapp,
 	}
 
 	tp_account_manager_set_all_requested_presences (tp_am, status, "", "");
+
+	g_object_unref (tp_am);
+}
+
+gboolean check_enabled_accounts ()
+{
+	gboolean r = FALSE;
+	TpAccountManager* tp_am = tp_account_manager_dup ();
+	GList* account_service_list = NULL;
+	GList* account_list = tp_account_manager_dup_valid_accounts (tp_am);
+
+	for (account_service_list = account_list; account_service_list != NULL; account_service_list = account_service_list->next) {
+		TpAccount* account = NULL;
+		account = (TpAccount*) account_service_list->data;
+		if (tp_account_is_enabled (account))
+		{
+			g_debug ("ENABLED: %s", tp_account_get_cm_name (account));
+			r = TRUE;
+		} else
+		{
+			g_debug ("DISABLED: %s", tp_account_get_cm_name (account));
+		}
+	}
+
+	g_list_free_full (account_list, g_object_unref);
+	g_object_unref(G_OBJECT(tp_am));
+	return r;
+}
+
+static void on_enabled_event (TpAccountManager* self, TpAccount* account, gpointer user_data)
+{
+	g_debug ("on_enabled_event");
+	if (mmapp)
+		messaging_menu_app_register (mmapp);
+}
+
+static void on_disabled_event (TpAccountManager* self, TpAccount* account, gpointer user_data)
+{
+	g_debug ("on_disabled_event");
+	if (mmapp && !check_enabled_accounts ())
+	{
+		messaging_menu_app_unregister (mmapp);
+	}
+}
+
+
+static void on_account_removed_event (TpAccountManager* self, TpAccount* account, gpointer user_data)
+{
+	g_debug ("on_account_removed_event");
+	if (mmapp && !check_enabled_accounts ())
+	{
+		messaging_menu_app_unregister (mmapp);
+	}
+}
+
+static void
+setup (GMainLoop *loop)
+{
+	TpAccountManager *tp_am;
+	TpBaseClient *approver;
+	TpBaseClient *observer;
+	GError *error = NULL;
+
+	g_debug (G_STRFUNC);
+
+	tp_am = tp_account_manager_dup ();
+
+	approver = approver_setup (tp_am, &error);
+	if (approver == NULL)
+	{
+		g_warning ("Failed to register the approver: %s", error->message);
+		g_error_free (error);
+		g_object_unref (tp_am);
+		g_main_loop_quit (loop);
+		return;
+	}
+
+	observer = observer_setup (tp_am, &error);
+	if (observer == NULL)
+	{
+		g_warning ("Failed to register the observer: %s", error->message);
+		g_error_free (error);
+		g_object_unref (tp_am);
+		g_main_loop_quit (loop);
+		return;
+	}
+
+	g_signal_connect(tp_am, "most-available-presence-changed",
+			 G_CALLBACK (presence_changed), NULL);
+	g_signal_connect (tp_am, "account-enabled", 
+			  G_CALLBACK (on_enabled_event), NULL);
+	g_signal_connect (tp_am, "account-disabled", 
+			  G_CALLBACK (on_disabled_event), NULL);
+	g_signal_connect (tp_am, "account-removed", 
+			  G_CALLBACK (on_account_removed_event), NULL);
+
+	dispatch_ops = g_hash_table_new_full (g_str_hash, g_str_equal,
+					      g_free, g_object_unref);
+	launcher = unity_launcher_entry_get_for_desktop_id ("empathy.desktop");
+
+	contact_list_setup (tp_am);
+
+	g_object_unref (tp_am);
+}
+
+static void
+free_ids_hash (gpointer data)
+{
+  GHashTable *ids = data;
+
+  if (mmapp)
+    {
+      GHashTableIter iter;
+      gpointer id;
+
+      g_hash_table_iter_init (&iter, ids);
+      while (g_hash_table_iter_next (&iter, &id, NULL))
+        {
+          messaging_menu_app_remove_source (mmapp, id);
+          --n_sources;
+        }
+
+      update_launcher (n_sources);
+    }
+
+  g_hash_table_unref (ids);
+}
+
+static void
+empathy_appeared (GDBusConnection *connection,
+		  const gchar     *name,
+		  const gchar     *name_owner,
+		  gpointer         user_data)
+{
+	TpAccountManager *tp_am;
+	TpConnectionPresenceType presence;
+
+	mmapp = messaging_menu_app_new ("empathy.desktop");
+	conns = g_hash_table_new_full (g_direct_hash, g_direct_equal, g_object_unref, free_ids_hash);
+
+	if (check_enabled_accounts ())
+		messaging_menu_app_register (mmapp);
+
+	g_signal_connect (mmapp, "activate-source",
+			  G_CALLBACK (message_source_activated), NULL);
+	g_signal_connect (mmapp, "status-changed",
+			  G_CALLBACK (messaging_menu_status_changed), NULL);
+
+	tp_am = tp_account_manager_dup ();
+	presence = tp_account_manager_get_most_available_presence (tp_am, NULL, NULL);
+	presence_changed (tp_am, presence, "", "", NULL);
+
+	g_object_unref (tp_am);
+}
+
+static void
+empathy_vanished (GDBusConnection *connection,
+		  const gchar     *name,
+		  gpointer         user_data)
+{
+	g_debug (G_STRFUNC);
+
+	g_clear_object (&launcher);
+	g_clear_object (&mmapp);
+	g_clear_pointer (&conns, g_hash_table_unref);
 }
 
 int
 main (int argc, char **argv)
 {
+	int ret = 0;
 	GMainLoop *loop = NULL;
-	TpConnectionPresenceType presence;
-
-	g_type_init ();
-
 	gtk_init (&argc, &argv);
-	TpBaseClient *approver;
-	TpBaseClient *observer;
 
 	loop = g_main_loop_new (NULL, FALSE);
+	setup (loop);
 
-	TpAccountManager *tp_am = tp_account_manager_dup ();
-
-	approver = approver_setup (tp_am);
-	if (approver == NULL)
-	{
-		g_error ("Approver registration failed, exiting");
-		return 1;
-	}
-	observer = observer_setup (tp_am);
-	if (observer == NULL)
-	{
-		g_error ("Observer registration failed, exiting");
-		return 1;
-	}
-
-	mmapp = messaging_menu_app_new ("empathy.desktop");
-	messaging_menu_app_register (mmapp);
-
-	g_signal_connect (mmapp, "activate-source",
-			  G_CALLBACK (message_source_activated), tp_am);
-	g_signal_connect (mmapp, "status-changed",
-			  G_CALLBACK (messaging_menu_status_changed), tp_am);
-
-	presence = tp_account_manager_get_most_available_presence (tp_am, NULL, NULL);
-	presence_changed (tp_am, presence, "", "", NULL);
-	g_signal_connect(tp_am, "most-available-presence-changed",
-			 G_CALLBACK (presence_changed), NULL);
-
-	dispatch_ops = g_hash_table_new_full (g_str_hash, g_str_equal,
-					      g_free, g_object_unref);
-
-	launcher = unity_launcher_entry_get_for_desktop_id ("empathy.desktop");
-
-	contact_list_setup (tp_am);
+	g_bus_watch_name (G_BUS_TYPE_SESSION,
+			  "org.gnome.Empathy",
+			  G_BUS_NAME_WATCHER_FLAGS_NONE,
+			  empathy_appeared,
+			  empathy_vanished,
+			  NULL, NULL);
 
 	g_debug ("Telepathy Indicator started");
 	g_main_loop_run (loop);
 
 	g_main_loop_unref (loop);
-	return 0;
+	return ret;
 }
